/*
PeRdr - PE file disassembler
Copyright (C) 1999-2006 Frediano Ziglio
-----

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
-----

INFORMATION
  www: http://freddy77.tripod.com/perdr/
  e-mail: freddy77@angelfire.com
*/
#include "global.h"
#ifdef HAVE_HDRSTOP
#pragma hdrstop
#endif

#include <cctype>

#include "module.h"
#include "code.h"
#include "codeglob.h"
#include "perdroptions.h"
#include "rawdump.h"

using namespace std;

// !!!
static bool commentBeginned = false;

// scrittura commenti
static void StartBeginComment()
{
  // scrivi inizio commento
  if (!commentBeginned)
    printf("\n");
  else
    printf("\n|\n");
  commentBeginned = true;

  printf("* ");
}

static void StartComment()
{
  if (!commentBeginned)
    StartBeginComment();
  else
    printf("\n|");
}

static void EndComment()
{
  if (commentBeginned)
  {
    printf("\n|\n");
    commentBeginned = false;
  }
}

// testa vari riferimenti
class CheckRef
{
public:
  CheckRef():count(0) {};
  void operator()(const IstrReference& ref)
  {
    if (!ref.direct)
    {
      char type = '?';
      switch(ref.type)
      {
      case FLOW_CALL:  if (options.showcRef) type = 'c'; break;
      case FLOW_JUMP:  if (options.showURef) type = 'U'; break;
      case FLOW_CJUMP: if (options.showCRef) type = 'C'; break;
      }
      if (type != '?')
      {
        if (count==0)
        {
          StartBeginComment();
          if (!options.showURef && !options.showCRef)
            printf("Referenced by a call at Address:");
          else
            printf("Referenced by an (U)nconditional or (C)onditional Jump "
                 "or (c)all at Address:");
        }
        if ( (count%4) == 0 )
          StartComment();
        else
          printf(",");
        if (!options.showURef && !options.showCRef)
          printf(" %08X",ref.to);
        else
          printf(" %08X(%c)",ref.to,type);
        ++count;
      }
    }
  }
  void Print()
  {
  }
private:
  int count;
};

void CodeParser::CheckReference(vma_t address)
{
  // entry point
  if (address == module->GetEntryPoint())
  {
    StartBeginComment();
    printf("Program entry point");
  }

  // export
  // !!! manca ordinal
  pair<Symbols::iterator,Symbols::iterator> sym = module->GetExportSymbols()[address];
  for(;sym.first != sym.second;++sym.first)
  {
    StartBeginComment();
    printf("Exported function %s",(*sym.first).name.c_str());
  }

  // testa presenza riferimento
  CheckRef refs;
  std::for_each(istrReferences.beginReferenceTo(address),
      istrReferences.endReferenceTo(address),
      refs);
  refs.Print();
}

#ifdef __BORLANDC__
#pragma argsused
#endif
void CodeParser::CheckJumpReference(vma_t address,const Instruction& instruction)
{
  // check for api reference
  for (int i=0; i<instruction.numArg; ++i)
  {
    const Param &arg = instruction.Args[i];
    if (arg.type == Param::t_memory)
    {
      // call/jmp dword ptr [constant]
      if (arg.mem_reg1 == null_reg && arg.mem_reg2 == null_reg)
      {
        pair<Symbols::iterator,Symbols::iterator> sym = module->GetSymbols()[arg.literal];
        for(;sym.first != sym.second;++sym.first)
        {
          StartBeginComment();
          printf("Reference to %s",(*sym.first).name.c_str());
        }
      }
    }
    if (arg.type == Param::t_literal)
    {
      // salti a jmp ad api
      if (IsPresent(apiAlias,arg.literal))
      {
        pair<Symbols::iterator,Symbols::iterator> sym = module->GetSymbols()[apiAlias[arg.literal]];
        for(;sym.first != sym.second;++sym.first)
        {
          StartBeginComment();
          printf("Reference to %s",(*sym.first).name.c_str());
        }
      }

      // salti a funzioni esportate
      pair<Symbols::iterator,Symbols::iterator> sym = module->GetExportSymbols()[arg.literal];
      for(;sym.first != sym.second;++sym.first)
      {
        StartBeginComment();
        printf("Reference to %s",(*sym.first).name.c_str());
      }
    }
  }
}

// !!! add code for other data, not only string
void CodeParser::CheckDataReference(vma_t address,const Instruction& instruction,std::string& comment)
{
  for (int i=0; i<instruction.numArg; ++i)
  {
    const Param &arg = instruction.Args[i];

    // if argument is a register ignore
    if (arg.type == Param::t_registry || arg.type == Param::t_farliteral)
      continue;
    
    vma_t address = arg.literal;
    if (arg.type == Param::t_memory)
    {
      if (!module->IsValid(address))
        continue;

      try
      {
        ObjectModule::DataReader reader = module->GetDataReader(address);
        address = reader.ReadU32LE();
      }
      catch(const OutOfAddress&)
      { continue; }
    }
    
    if (!module->IsValid(address))
      continue;

    unsigned len = byteInfo.GetLen(address);
    ByteInfo::TType type = byteInfo[address].GetType();
    if (type == ByteInfo::typeAnsiZ || type == ByteInfo::typeUnicodeZ ||
        IsPossibleString(address, len, type))
    {
      StartBeginComment();
      printf("Possible string reference %04Xh ",address);
      WriteString(address, len, type, 100);
    }
  }
}

// stampa una stringa
void CodeParser::WriteBinary(vma_t address,int len) const
{
  EndComment();

  _PRG_ASSERT(&*module != NULL);
  _PRG_ASSERT(len>0);
  ObjectModule::DataReader reader = module->GetDataReader(address);
  uint8_t buffer[256];
  while (len > 0)
  {
    unsigned left = (len > 256) ? 256u : (unsigned) len;
    for (unsigned n = 0; n < left; ++n)
      buffer[n] = reader.ReadU8();
    ::WriteBinary(address,buffer,left);
    address += left;
    len -= left;
  }
}

// scrive inizio istruzione/BYTE/DWORD come da inizio
// legge da ::module
void CodeParser::WriteBegin(vma_t address,int num_bytes,bool truncate) const
{
  int x=0,c;
  EndComment();
  printf("%08lX ",(long int)address);
  if (::options.showBytes)
  {
    if (truncate)
    {
      if (num_bytes > 10)
        num_bytes = 9;
      else
        truncate = false;
    }
    ObjectModule::DataReader reader = module->GetDataReader(address);
    for (int i=0;i<num_bytes;++i)
    {
      if ( x == 10 && num_bytes != 11)
      {
        x = 0;
        if (i != 0)
          printf("\n         ");
      }
      c = reader.ReadU8();
      printf("%02X",c);
      ++x;
    }

    if (truncate)
      printf("..");
    else for (;x<10;++x)
      printf("  ");
  }

  printf("\t");
}

// output a string constant
int CodeParser::WriteString(vma_t address, int len, ByteInfo::TType type, int truncate_at) const
{
  ObjectModule::DataReader reader = module->GetDataReader(address);
  bool uc = (type == ByteInfo::typeUnicodeZ);
  if (uc) {
    printf("L");
    len /= 2;
  }
  printf("\"");
  // let's not go overboard...
  const char* truncated = "";
  if (truncate_at > 0 && len > truncate_at)
  {
    len = truncate_at;
    truncated = "...";
  }
  while (--len > 0)
  {
    int c = reader.ReadU8();
    if (uc)
      c |= reader.ReadU8() << 8;
    switch(c)
    {
    case '\t':
      printf("\\t");
      break;
    case '\r':
      printf("\\r");
      break;
    case '\n':
      printf("\\n");
      break;
    case '\"':
      printf("\\\"");
      break;
    default:
      if (c < 256 && isprint(c))
        printf("%c",c);
      else if (c < 256)
        printf("\\x%02x",c);
      else
        printf("\\u%04x", c);
    }
  }
  printf("\"%s", truncated);

  return (uc) ? 2 : 1;
}

static void WritePriority(ByteInfo::TPriority priority)
{
  const char *s = "unknown";
  switch (priority)
  {
  case ByteInfo::priNone:
    s = "none"; break;
  case ByteInfo::priCheckOnly:
    s = "check only"; break;
  case ByteInfo::priFiller:
    s = "filler"; break;
  case ByteInfo::priConstant:
    s = "constant"; break;
  case ByteInfo::priApi:
    s = "api"; break;
  case ByteInfo::priExport:
    s = "export"; break;
  case ByteInfo::priHeuristics:
    s = "heuristics"; break;
  case ByteInfo::priSafeHeuristics:
    s = "safe heuristics"; break;
  case ByteInfo::priSafeExport:
    s = "safe export"; break;
  case ByteInfo::priEntryPoint:
    s = "entry point"; break;
  }
  StartBeginComment();
  printf("Current priority: %s",s);
}

void CodeParser::WriteCode(vma_t writeStart, vma_t writeEnd)
{
  vma_t end;

  // disassembly vero e proprio
  FOR_EACH_SECTION_CODE_BEGIN(module,p)

    InstructionDecoder decoder(module->GetRelocationInfos());

    // init as invalid priority
    ByteInfo::TPriority priority = ByteInfo::TPriority(-1);

    vma_t pc = (*p).begin;
    end      = (*p).end;

    if (writeStart || writeEnd)
    {
      if (pc < writeStart)
        pc = writeStart;
      if (end > writeEnd)
        end = writeEnd;
    }

    for(; pc<end; )
    {
      ObjectModule::DataReader reader = module->GetDataReader(pc);
      int c;

      // testa cambio priorita
      if (options.showPriority)
      {
        ByteInfo::TPriority currPriority = byteInfo[pc].GetPriority();
        if (currPriority != priority)
        {
          WritePriority(currPriority);
          priority = currPriority;
        }
      }

      // test for reference
      CheckReference(pc);

      // testa se tipo conosciuto
      unsigned len = byteInfo.GetLen(pc);
      bool isInstruction = true;

      if (len != 0)
      {
        isInstruction = false;
        const ByteInfo& info = byteInfo[pc];
        switch (info.GetType())
        {
        case ByteInfo::typeInteger:
          {
            uint32_t i;
            switch (len)
            {
            case 1:
              i = reader.ReadU8();
              WriteBegin(pc,len);
              printf("BYTE\t%02X\n",i);
              break;
            case 2:
              i = reader.ReadU16LE();
              WriteBegin(pc,len);
              printf("WORD\t%04X\n",i);
              break;
            case 4:
              i = reader.ReadU32LE();
              WriteBegin(pc,len);
              printf("DWORD\t%08X\n",i);
              break;

            default:
              // !!! scrive come bytes normali
              //_PRG_ASSERT(0);
              WriteBinary(pc,len);
            }
            pc += len;
          }
          break;
        case ByteInfo::typeFloat:
          {
            // !!! not that portable
            union {
              struct {
                uint32_t i4;
                uint32_t i8;
                uint16_t i10;
              };
              float f4;
              double f8;
              long double f10;
            } i;
            switch (len)
            {
            case 4:
              i.i4 = reader.ReadU32LE();
              WriteBegin(pc,len);
              // !!! precision is sufficient ??
              printf("DWORD\t%#.7g\n",i.f4);
              break;
            case 8:
              i.i4 = reader.ReadU32LE();
              i.i8 = reader.ReadU32LE();
              WriteBegin(pc,len);
              // !!! precision is sufficient ??
              printf("QWORD\t%#.7g\n",i.f8);
              break;
            case 10:
              i.i4 = reader.ReadU32LE();
              i.i8 = reader.ReadU32LE();
              i.i10 = reader.ReadU16LE();
              WriteBegin(pc,len);
              // !!! precision is sufficient ??
              printf("TWORD\t%#.7Lg\n",i.f10);
              break;

            default:
              // !!! scrive come bytes normali
              //_PRG_ASSERT(0);
              WriteBinary(pc,len);
            }
            pc += len;
          }
          break;
        case ByteInfo::typePointer:
          {
            _PRG_ASSERT(len == (unsigned int)addr_bytes);
            vma_t address = reader.ReadU32LE();
            WriteBegin(pc,len);
            printf("DWORD\t%08lX\n",(long int)address);
            pc += len;
          }
          break;
        case ByteInfo::typeLoader:
          StartBeginComment();
          printf("Loader bytes");
          WriteBinary(pc,len);
          pc += len;
          break;
        case ByteInfo::typeAnsiZ:
        case ByteInfo::typeUnicodeZ:
          WriteBegin(pc, len, true);
          printf("StringZ\t");
          WriteString(pc, len, info.GetType());
          printf("\n");
          pc += len;
          break;
        default:
          isInstruction = true;
        }
      }

      if (isInstruction)
      {
        Instruction currInstruction;

        // read next instruction
        int result = decoder.Decode(currInstruction,reader);

        if ( !result || (byteInfo.GetIstrLen(pc)==0 &&
             byteInfo.IsOccupied(pc,result)) )
        {
          reader = module->GetDataReader(pc);
          c = reader.ReadU8();
          WriteBegin(pc,1);
          printf("BYTE\t%02X\n",c);
          pc += 1;
        }
        else
        {
          std::string comment;
          CheckJumpReference(pc,currInstruction);
          CheckDataReference(pc, currInstruction, comment);
          WriteBegin(pc,result);
          WriteInstruction(currInstruction, comment);
          // go on
          pc += result;
        }
      }
    }
  FOR_EACH_SECTION_CODE_END(module,p)
}

// vim:tabstop=200:mouse=a:
